import ast
import sqlite3

from fastapi import FastAPI, status
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles

from typing import List
from pydantic import BaseModel

API_BASE_PATH = "/api"

# Set up a connection to the database.
db = sqlite3.connect("movies.db")
db_cur = db.cursor()

app = FastAPI()

origins = [
    # Allow all localhost.
    "http://localhost",
    "http://localhost:63344",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


def custom_openapi():
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="CinemaSeek",
        version="1.0.0",
        description="An API about movie data.",
        routes=app.routes,
    )
    openapi_schema["info"]["x-logo"] = {
        "url": "https://fastapi.tiangolo.com/img/logo-margin/logo-teal.png"
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi
app.mount("/ui", StaticFiles(directory="frontend"), name="frontend")


# Wrote these models for the sake of making examples appear in /doc and /redoc.
class GenericResponse(BaseModel):
    status: str


class ActorHighGrossing(GenericResponse):
    actor: str
    movies: dict[str, int]


class FeaturePercent(GenericResponse):
    actor: str
    percentage: float


class DirectorAsActor(GenericResponse):
    movie: str
    director_as_actor: str


class StudioMovieList(GenericResponse):
    studio: str
    movies: List[str]


class HighGrossLowRate(GenericResponse):
    movies: dict[str, float]


class YearlyAverages(GenericResponse):
    years: dict[str, int]


class FlixAndTheatre(GenericResponse):
    movie: str
    in_theatre: bool
    in_netflix: bool


def make_full_name(first, last):
    """Get a full name string from the provided first and last name."""
    first = first.lower()
    first = first[0].upper() + first[1:]
    last = last.lower()
    last = last[0].upper() + last[1:]
    return first + ' ' + last


# API #
@app.get(f"{API_BASE_PATH}/actor/high-grossing-movies", response_model=ActorHighGrossing)
async def get_actor_high_grossing_movies(first_name: str, last_name: str):
    """Return a list of high grossing films the provided actor is in ordered by revenue."""
    first_name = first_name.strip()
    last_name = last_name.strip()
    full_name = make_full_name(first_name, last_name)
    movie_ids_raw = list(db_cur.execute(f"SELECT `id` FROM `movies_credits`"
                                        f"WHERE `cast` LIKE '%''name'': ''{full_name}''%';"))
    movie_ids = tuple([int(movie_id[0]) for movie_id in movie_ids_raw])
    if not movie_ids:
        response = {"status": "Actor not found!", "actor": full_name}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=response)

    # If-else added to prevent an SQL formatting error in case of 1 movie id.
    if len(movie_ids) > 1:
        movie_names_raw = list(db_cur.execute(f"SELECT `original_title` FROM `movies_metadata`"
                                              f"WHERE `id` IN {movie_ids};"))
    else:
        movie_names_raw = list(db_cur.execute(f"SELECT `original_title` FROM `movies_metadata`"
                                              f"WHERE `id` = {movie_ids[0]};"))
    movie_names = [movie_name[0] for movie_name in movie_names_raw]

    title_check_queries = ''
    for i in range(len(movie_names)):
        title_check_queries += f'`Title` LIKE "%{movie_names[i]}%"'
        if i < len(movie_names) - 1:
            title_check_queries += " OR "
    high_grossers_raw = list(db_cur.execute(f"SELECT substr(`Title`, 1, length(`Title`) - 7), `World Sales (in $)` "
                                            f"FROM `highest_grossing_movies`"
                                            f"WHERE {title_check_queries} "
                                            f"ORDER BY `World Sales (in $)` DESC;"))
    high_grossers = {}
    for movie in high_grossers_raw:
        high_grossers[movie[0]] = movie[1]
    response = {"status": "Actor found!", "actor": full_name, "movies": high_grossers}
    return JSONResponse(status_code=status.HTTP_200_OK, content=response)


@app.get(f"{API_BASE_PATH}/actor/feature-percent", response_model=FeaturePercent)
async def get_feature_percent(first_name: str, last_name: str):
    """Return the feature percentage of an actor rounded to four decimal places."""
    first_name = first_name.strip()
    last_name = last_name.strip()
    full_name = make_full_name(first_name, last_name)
    appearance_count = list(db_cur.execute(f"SELECT COUNT(`cast`) FROM `movies_credits`"
                                           f"WHERE `cast` LIKE '%''name'': ''{full_name}''%';"))[0][0]
    if not appearance_count:
        response = {"status": "Actor not found!", "actor": full_name}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=response)
    movie_count = list(db_cur.execute(f"SELECT COUNT(`cast`) FROM `movies_credits`"))[0][0]
    appearance_percentage = appearance_count / movie_count * 100
    print(appearance_count)
    response = {"status": "Actor found!", "actor": full_name, "percentage": round(appearance_percentage, 4)}
    return JSONResponse(status_code=status.HTTP_200_OK, content=response)


@app.get(f"{API_BASE_PATH}/movie/director-as-actor", response_model=DirectorAsActor)
async def get_director_as_actor(movie: str):
    """Input a movie name and the name of the director who doubles as an actor will be returned."""
    movie = movie.strip()
    movie_id_raw = list(db_cur.execute(f"SELECT `id` FROM `movies_metadata`"
                                       f"WHERE `original_title` = '{movie}'"))
    if not movie_id_raw:
        response = {"status": "Movie not found!", "movie": movie}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=response)
    movie_id = movie_id_raw[0][0]

    casts_n_crews_raw = list(db_cur.execute(f"SELECT `cast`, `crew` FROM `movies_credits`"
                                            f"WHERE `id` = {movie_id}"))
    casts_n_crews = [ast.literal_eval(casts_n_crews_raw[0][0]), ast.literal_eval(casts_n_crews_raw[0][1])]
    director = ''
    for crew in casts_n_crews[1]:
        if crew["job"] == "Director":
            director = crew["name"]
            break
    for cast in casts_n_crews[0]:
        if cast["name"] == director:
            response = {"status": "Movie found!", "movie": movie, "director_as_actor": director}
            return JSONResponse(status_code=status.HTTP_200_OK, content=response)
    response = {"status": "Movie found!", "movie": movie, "director_as_actor": None}
    return JSONResponse(status_code=status.HTTP_200_OK, content=response)


@app.get(f"{API_BASE_PATH}/movie/studio-movie-list", response_model=StudioMovieList)
async def get_studio_movie_list(studio: str):
    """Take studio name as a param and return every movie the studio has ever made."""
    studio = studio.strip()
    movies_raw = list(db_cur.execute(f"SELECT original_title FROM `movies_metadata` "
                                     f"WHERE production_companies LIKE '%''{studio}''%'"
                                     f"ORDER BY original_title"))
    movies = [movie[0] for movie in movies_raw]
    if not movies:
        response = {"status": "Studio not found!", "studio": studio}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=response)
    response = {"status": "Studio found!", "studio": studio, "movies": movies}
    return JSONResponse(status_code=status.HTTP_200_OK, content=response)


@app.get(f"{API_BASE_PATH}/rating/high-gross-low-rate", response_model=HighGrossLowRate)
async def get_high_gross_low_rate():
    """List high grossing movies which has rating lower than 5 (scale of 10). Ratings provided by MovieLens."""
    month_in_double_digits = {
        "January": "01",
        "February": "02",
        "March": "03",
        "April": "04",
        "May": "05",
        "June": "06",
        "July": "07",
        "August": "08",
        "September": "09",
        "October": "10",
        "November": "11",
        "December": "12"
    }

    high_grossers_raw = list(db_cur.execute("SELECT SUBSTR(`Title`, 1, LENGTH(`Title`) - 7), `Release Date` "
                                            "FROM `highest_grossing_movies`;"))
    high_grossers = []
    for movie in high_grossers_raw:
        title = movie[0]
        date = movie[1]

        if date != "NA":
            date_parts = movie[1].split(' ')
            year = date_parts[2]
            month = month_in_double_digits[date_parts[0]]
            date = date_parts[1][:-1]  # Remove comma from date.
            date = f"{year}-{month}-{date}"

        high_grossers.append([title, date])

    # Partition request to prevent parser stack overflow.
    high_gross_low_rates_raw = []
    partition = []
    partition_size = len(high_grossers) / 6
    for i in range(len(high_grossers)):
        partition.append(high_grossers[i])
        if len(partition) == partition_size or i == len(high_grossers) - 1:
            title_check_queries = ''
            for j in range(len(partition)):
                title = partition[j][0]
                date = partition[j][1]
                if date != "NA":
                    # Need to check release date as well in case of movies with the same name.
                    title_check_queries += f'(`original_title` = "{title}" AND `release_date` = "{date}")'
                else:
                    title_check_queries += f'(`original_title` = "{title}")'
                if j < len(partition) - 1:
                    title_check_queries += " OR "

            high_gross_low_rates_raw.extend(list(
                db_cur.execute(f"SELECT `original_title`, `vote_average` FROM `movies_metadata`"
                               f"WHERE ({title_check_queries}) AND `vote_average` < 5 AND `vote_average` > 0 "
                               f"ORDER BY `vote_average` DESC;")))
            partition.clear()

    high_gross_low_rates = {}
    for movie in high_gross_low_rates_raw:
        high_gross_low_rates[movie[0]] = movie[1]
    response = {"status": "Success!", "movies": high_gross_low_rates}
    return JSONResponse(status_code=status.HTTP_200_OK, content=response)


@app.get(f"{API_BASE_PATH}/rating/yearly-averages", response_model=YearlyAverages)
async def get_yearly_averages():
    """Return a list of average movie rating (scale of 10) for every available year. Ratings provided by MovieLens."""
    yearly_ratings_raw = list(db_cur.execute(f"SELECT SUBSTR(`release_date`, 1, 4) AS year,"
                                             f"AVG(`vote_average`) AS avg_rating FROM `movies_metadata` "
                                             f"GROUP BY SUBSTR(`release_date`, 1, 4)"))
    yearly_ratings = {}
    for yearly_rating in yearly_ratings_raw:
        if yearly_rating[0] and yearly_rating[1] > 0:
            # No invalid years or ratings will be included.
            yearly_ratings[yearly_rating[0]] = yearly_rating[1]
    response = {"status": "Yearly averages calculated!", "years": yearly_ratings}
    return JSONResponse(status_code=status.HTTP_200_OK, content=response)


@app.get(f"{API_BASE_PATH}/platform/flix-and-theatre", response_model=FlixAndTheatre)
async def get_flix_and_theatre(movie: str):
    """Input a movie name and return whether it is in theatre and whether it is in Netflix."""
    movie = movie.strip()
    in_theatre = bool(list(db_cur.execute(f"SELECT `original_title` FROM `movies_metadata`"
                                          f"WHERE `original_title` = '{movie}'")))
    in_netflix = bool(list(db_cur.execute(f"SELECT `title` FROM `netflix`"
                                          f"WHERE `title` = '{movie}' AND `type` = 'MOVIE'")))
    if not in_theatre and not in_netflix:
        response = {"status": "Movie not found, or the provided input is a show!", "movie": movie}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=response)
    response = {"status": "Movie found!", "movie": movie, "in_theatre": in_theatre, "in_netflix": in_netflix}
    return JSONResponse(status_code=status.HTTP_200_OK, content=response)
